@@ -1,9 +1,9 @@
+from dagster import Field, Float, Noneable, StringSource
+from dagster.core.host_representation import IN_PROCESS_NAME
+from dagster.utils import merge_dicts
 from dagster_celery.executor import CELERY_CONFIG
 from dagster_k8s import DagsterK8sJobConfig
-
-from dagster import Field, Noneable, StringSource
-from dagster.core.host_representation.handle import IN_PROCESS_NAME
-from dagster.utils import merge_dicts
+from dagster_k8s.client import DEFAULT_WAIT_TIMEOUT
 
 CELERY_K8S_CONFIG_KEY = "celery-k8s"
 
@@ -11,7 +11,9 @@ CELERY_K8S_CONFIG_KEY = "celery-k8s"
 def celery_k8s_config():
 
     # DagsterK8sJobConfig provides config schema for specifying Dagster K8s Jobs
-    job_config = DagsterK8sJobConfig.config_type_pipeline_run()
+    job_config = DagsterK8sJobConfig.config_type_pipeline_run(
+        default_image_pull_policy="IfNotPresent"
+    )
 
     additional_config = {
         "load_incluster_config": Field(
@@ -43,6 +45,12 @@ def celery_k8s_config():
             default_value=IN_PROCESS_NAME,
             description="The repository location name to use for execution.",
         ),
+        "job_wait_timeout": Field(
+            Float,
+            is_required=False,
+            default_value=DEFAULT_WAIT_TIMEOUT,
+            description=f"Wait this many seconds for a job to complete before marking the run as failed. Defaults to {DEFAULT_WAIT_TIMEOUT} seconds.",
+        ),
     }
 
     cfg = merge_dicts(CELERY_CONFIG, job_config)
@@ -50,8 +58,7 @@ def celery_k8s_config():
     return cfg
 
 
-def get_celery_engine_config():
-
+def get_celery_engine_config(additional_env_config_maps=None):
     return {
         "execution": {
             CELERY_K8S_CONFIG_KEY: {
@@ -59,7 +66,12 @@ def get_celery_engine_config():
                     "job_image": {"env": "DAGSTER_K8S_PIPELINE_RUN_IMAGE"},
                     "job_namespace": {"env": "DAGSTER_K8S_PIPELINE_RUN_NAMESPACE"},
                     "image_pull_policy": {"env": "DAGSTER_K8S_PIPELINE_RUN_IMAGE_PULL_POLICY"},
-                    "env_config_maps": [{"env": "DAGSTER_K8S_PIPELINE_RUN_ENV_CONFIGMAP"}],
+                    "env_config_maps": (
+                        [
+                            {"env": "DAGSTER_K8S_PIPELINE_RUN_ENV_CONFIGMAP"},
+                        ]
+                        + (additional_env_config_maps if additional_env_config_maps else [])
+                    ),
                 }
             }
         }
