@@ -1,11 +1,39 @@
+import logging
 import os
+import sys
+import time
 
 import kubernetes
+from dagster import (
+    DagsterEvent,
+    DagsterEventType,
+    DagsterInstance,
+    EventMetadataEntry,
+    Executor,
+    check,
+    executor,
+    multiple_process_executor_requirements,
+)
+from dagster.cli.api import ExecuteStepArgs
+from dagster.core.errors import DagsterUnmetExecutorRequirementsError
+from dagster.core.events import EngineEventData
+from dagster.core.events.log import EventLogEntry
+from dagster.core.execution.plan.objects import StepFailureData, UserFailureData
+from dagster.core.execution.retries import RetryMode
+from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
+from dagster.serdes import pack_value, serialize_dagster_namedtuple, unpack_value
+from dagster.utils.error import serializable_error_info_from_exc_info
 from dagster_celery.config import DEFAULT_CONFIG, dict_wrapper
 from dagster_celery.core_execution_loop import DELEGATE_MARKER
 from dagster_celery.defaults import broker_url, result_backend
 from dagster_k8s import DagsterK8sJobConfig, construct_dagster_k8s_job
-from dagster_k8s.client import DagsterK8sPipelineStatusException
+from dagster_k8s.client import (
+    DagsterK8sAPIRetryLimitExceeded,
+    DagsterK8sError,
+    DagsterK8sPipelineStatusException,
+    DagsterK8sTimeoutError,
+    DagsterK8sUnrecoverableAPIError,
+)
 from dagster_k8s.job import (
     UserDefinedDagsterK8sConfig,
     get_k8s_job_name,
@@ -19,21 +47,15 @@ from dagster_k8s.utils import (
     wait_for_job_success,
 )
 
-from dagster import DagsterInstance, EventMetadataEntry, Executor, check, executor
-from dagster.cli.api import ExecuteStepArgs
-from dagster.core.definitions.executor import check_cross_process_constraints
-from dagster.core.events import EngineEventData
-from dagster.core.execution.retries import Retries
-from dagster.core.instance import InstanceRef
-from dagster.core.origin import PipelineOrigin
-from dagster.core.storage.pipeline_run import PipelineRunStatus
-from dagster.serdes import pack_value, serialize_dagster_namedtuple, unpack_value
-
 from .config import CELERY_K8S_CONFIG_KEY, celery_k8s_config
 from .launcher import CeleryK8sRunLauncher
 
 
-@executor(name=CELERY_K8S_CONFIG_KEY, config_schema=celery_k8s_config())
+@executor(
+    name=CELERY_K8S_CONFIG_KEY,
+    config_schema=celery_k8s_config(),
+    requirements=multiple_process_executor_requirements(),
+)
 def celery_k8s_job_executor(init_context):
     """Celery-based executor which launches tasks as Kubernetes Jobs.
 
@@ -48,15 +70,13 @@ def celery_k8s_job_executor(init_context):
 
     In the most common case, you may want to modify the ``broker`` and ``backend`` (e.g., to use
     Redis instead of RabbitMQ). We expect that ``config_source`` will be less frequently
-    modified, but that when solid executions are especially fast or slow, or when there are
-    different requirements around idempotence or retry, it may make sense to execute pipelines
+    modified, but that when op executions are especially fast or slow, or when there are
+    different requirements around idempotence or retry, it may make sense to execute dagster jobs
     with variations on these settings.
 
-    If you'd like to configure a Celery Kubernetes Job executor in addition to the
-    :py:class:`~dagster.default_executors`, you should add it to the ``executor_defs`` defined on a
-    :py:class:`~dagster.ModeDefinition` as follows:
+    To use the `celery_k8s_job_executor`, set it as the `executor_def` when defining a job:
 
-    .. literalinclude:: ../dagster_celery_k8s_tests/example_celery_mode_def.py
+    .. literalinclude:: ../../../../../../python_modules/libraries/dagster-celery-k8s/dagster_celery_k8s_tests/example_celery_mode_def.py
        :language: python
 
     Then you can configure the executor as follows:
@@ -64,16 +84,15 @@ def celery_k8s_job_executor(init_context):
     .. code-block:: YAML
 
         execution:
-          celery-k8s:
-            config:
-              job_image: 'my_repo.com/image_name:latest'
-              job_namespace: 'some-namespace'
-              broker: 'pyamqp://guest@localhost//'  # Optional[str]: The URL of the Celery broker
-              backend: 'rpc://' # Optional[str]: The URL of the Celery results backend
-              include: ['my_module'] # Optional[List[str]]: Modules every worker should import
-              config_source: # Dict[str, Any]: Any additional parameters to pass to the
-                  #...       # Celery workers. This dict will be passed as the `config_source`
-                  #...       # argument of celery.Celery().
+          config:
+            job_image: 'my_repo.com/image_name:latest'
+            job_namespace: 'some-namespace'
+            broker: 'pyamqp://guest@localhost//'  # Optional[str]: The URL of the Celery broker
+            backend: 'rpc://' # Optional[str]: The URL of the Celery results backend
+            include: ['my_module'] # Optional[List[str]]: Modules every worker should import
+            config_source: # Dict[str, Any]: Any additional parameters to pass to the
+                #...       # Celery workers. This dict will be passed as the `config_source`
+                #...       # argument of celery.Celery().
 
     Note that the YAML you provide here must align with the configuration with which the Celery
     workers on which you hope to run were started. If, for example, you point the executor at a
@@ -84,17 +103,14 @@ def celery_k8s_job_executor(init_context):
     commands must be invoked with the `-A dagster_celery_k8s.app` argument.
     """
 
-    check_cross_process_constraints(init_context)
-
     run_launcher = init_context.instance.run_launcher
     exc_cfg = init_context.executor_config
 
-    check.inst(
-        run_launcher,
-        CeleryK8sRunLauncher,
-        "This engine is only compatible with a CeleryK8sRunLauncher; configure the "
-        "CeleryK8sRunLauncher on your instance to use it.",
-    )
+    if not isinstance(run_launcher, CeleryK8sRunLauncher):
+        raise DagsterUnmetExecutorRequirementsError(
+            "This engine is only compatible with a CeleryK8sRunLauncher; configure the "
+            "CeleryK8sRunLauncher on your instance to use it.",
+        )
 
     job_config = DagsterK8sJobConfig(
         dagster_home=run_launcher.dagster_home,
@@ -106,6 +122,8 @@ def celery_k8s_job_executor(init_context):
         service_account_name=exc_cfg.get("service_account_name"),
         env_config_maps=exc_cfg.get("env_config_maps"),
         env_secrets=exc_cfg.get("env_secrets"),
+        volume_mounts=exc_cfg.get("volume_mounts"),
+        volumes=exc_cfg.get("volumes"),
     )
 
     # Set on the instance but overrideable here
@@ -113,7 +131,7 @@ def celery_k8s_job_executor(init_context):
     backend = run_launcher.backend or exc_cfg.get("backend")
     config_source = run_launcher.config_source or exc_cfg.get("config_source")
     include = run_launcher.include or exc_cfg.get("include")
-    retries = run_launcher.retries or Retries.from_config(exc_cfg.get("retries"))
+    retries = run_launcher.retries or RetryMode.from_config(exc_cfg.get("retries"))
 
     return CeleryK8sJobExecutor(
         broker=broker,
@@ -126,6 +144,7 @@ def celery_k8s_job_executor(init_context):
         load_incluster_config=exc_cfg.get("load_incluster_config"),
         kubeconfig_file=exc_cfg.get("kubeconfig_file"),
         repo_location_name=exc_cfg.get("repo_location_name"),
+        job_wait_timeout=exc_cfg.get("job_wait_timeout"),
     )
 
 
@@ -142,6 +161,7 @@ class CeleryK8sJobExecutor(Executor):
         load_incluster_config=False,
         kubeconfig_file=None,
         repo_location_name=None,
+        job_wait_timeout=None,
     ):
 
         if load_incluster_config:
@@ -152,7 +172,7 @@ class CeleryK8sJobExecutor(Executor):
         else:
             check.opt_str_param(kubeconfig_file, "kubeconfig_file")
 
-        self._retries = check.inst_param(retries, "retries", Retries)
+        self._retries = check.inst_param(retries, "retries", RetryMode)
         self.broker = check.opt_str_param(broker, "broker", default=broker_url)
         self.backend = check.opt_str_param(backend, "backend", default=result_backend)
         self.include = check.opt_list_param(include, "include", of_type=str)
@@ -167,7 +187,8 @@ class CeleryK8sJobExecutor(Executor):
         )
 
         self.kubeconfig_file = check.opt_str_param(kubeconfig_file, "kubeconfig_file")
-        self.repo_location_name = check.str_param(repo_location_name, "repo_location_name")
+        self.repo_location_name = check.opt_str_param(repo_location_name, "repo_location_name")
+        self.job_wait_timeout = check.float_param(job_wait_timeout, "job_wait_timeout")
 
     @property
     def retries(self):
@@ -190,27 +211,36 @@ class CeleryK8sJobExecutor(Executor):
         }
 
 
-def _submit_task_k8s_job(app, pipeline_context, step, queue, priority):
+def _submit_task_k8s_job(app, pipeline_context, step, queue, priority, known_state):
     user_defined_k8s_config = get_user_defined_k8s_config(step.tags)
 
+    pipeline_origin = pipeline_context.reconstructable_pipeline.get_python_origin()
+
+    execute_step_args = ExecuteStepArgs(
+        pipeline_origin=pipeline_origin,
+        pipeline_run_id=pipeline_context.pipeline_run.run_id,
+        step_keys_to_execute=[step.key],
+        instance_ref=pipeline_context.instance.get_ref(),
+        retry_mode=pipeline_context.executor.retries.for_inner_plan(),
+        known_state=known_state,
+        should_verify_step=True,
+    )
+
+    job_config = pipeline_context.executor.job_config
+    if not job_config.job_image:
+        job_config = job_config.with_image(pipeline_origin.repository_origin.container_image)
+
+    if not job_config.job_image:
+        raise Exception("No image included in either executor config or the dagster job")
+
     task = create_k8s_job_task(app)
-
-    recon_repo = pipeline_context.pipeline.get_reconstructable_repository()
-
     task_signature = task.si(
-        instance_ref_dict=pipeline_context.instance.get_ref().to_dict(),
-        step_keys=[step.key],
-        run_config=pipeline_context.pipeline_run.run_config,
-        mode=pipeline_context.pipeline_run.mode,
-        repo_name=recon_repo.get_definition().name,
-        repo_location_name=pipeline_context.executor.repo_location_name,
-        run_id=pipeline_context.pipeline_run.run_id,
-        job_config_dict=pipeline_context.executor.job_config.to_dict(),
+        execute_step_args_packed=pack_value(execute_step_args),
+        job_config_dict=job_config.to_dict(),
         job_namespace=pipeline_context.executor.job_namespace,
         user_defined_k8s_config_dict=user_defined_k8s_config.to_dict(),
-        retries_dict=pipeline_context.executor.retries.for_inner_plan().to_config(),
-        pipeline_origin_packed=pack_value(pipeline_context.pipeline.get_origin()),
         load_incluster_config=pipeline_context.executor.load_incluster_config,
+        job_wait_timeout=pipeline_context.executor.job_wait_timeout,
         kubeconfig_file=pipeline_context.executor.kubeconfig_file,
     )
 
@@ -221,38 +251,55 @@ def _submit_task_k8s_job(app, pipeline_context, step, queue, priority):
     )
 
 
+def construct_step_failure_event_and_handle(pipeline_run, step_key, err, instance):
+    step_failure_event = DagsterEvent(
+        event_type_value=DagsterEventType.STEP_FAILURE.value,
+        pipeline_name=pipeline_run.pipeline_name,
+        step_key=step_key,
+        event_specific_data=StepFailureData(
+            error=serializable_error_info_from_exc_info(sys.exc_info()),
+            user_failure_data=UserFailureData(label="K8sError"),
+        ),
+    )
+    event_record = EventLogEntry(
+        message=str(err),
+        user_message=str(err),
+        level=logging.ERROR,
+        pipeline_name=pipeline_run.pipeline_name,
+        run_id=pipeline_run.run_id,
+        error_info=None,
+        step_key=step_key,
+        timestamp=time.time(),
+        dagster_event=step_failure_event,
+    )
+    instance.handle_new_event(event_record)
+    return step_failure_event
+
+
 def create_k8s_job_task(celery_app, **task_kwargs):
     @celery_app.task(bind=True, name="execute_step_k8s_job", **task_kwargs)
     def _execute_step_k8s_job(
         self,
-        instance_ref_dict,
-        step_keys,
-        run_config,
-        mode,
-        repo_name,
-        repo_location_name,
-        run_id,
+        execute_step_args_packed,
         job_config_dict,
         job_namespace,
         load_incluster_config,
-        retries_dict,
-        pipeline_origin_packed,
+        job_wait_timeout,
         user_defined_k8s_config_dict=None,
         kubeconfig_file=None,
     ):
-        """Run step execution in a K8s job pod.
-        """
-
-        check.dict_param(instance_ref_dict, "instance_ref_dict")
-        check.list_param(step_keys, "step_keys", of_type=str)
+        """Run step execution in a K8s job pod."""
+        execute_step_args = unpack_value(
+            check.dict_param(
+                execute_step_args_packed,
+                "execute_step_args_packed",
+            )
+        )
+        check.inst_param(execute_step_args, "execute_step_args", ExecuteStepArgs)
         check.invariant(
-            len(step_keys) == 1, "Celery K8s task executor can only execute 1 step at a time"
+            len(execute_step_args.step_keys_to_execute) == 1,
+            "Celery K8s task executor can only execute 1 step at a time",
         )
-        check.dict_param(run_config, "run_config")
-        check.str_param(mode, "mode")
-        check.str_param(repo_name, "repo_name")
-        check.str_param(repo_location_name, "repo_location_name")
-        check.str_param(run_id, "run_id")
 
         # Celery will serialize this as a list
         job_config = DagsterK8sJobConfig.from_dict(job_config_dict)
@@ -260,20 +307,14 @@ def create_k8s_job_task(celery_app, **task_kwargs):
         check.str_param(job_namespace, "job_namespace")
 
         check.bool_param(load_incluster_config, "load_incluster_config")
-        check.dict_param(retries_dict, "retries_dict")
-
-        pipeline_origin = unpack_value(
-            check.dict_param(
-                pipeline_origin_packed, "pipeline_origin_packed"
-            )  # TODO: make part of args
-        )
-        check.inst(pipeline_origin, PipelineOrigin)
 
         user_defined_k8s_config = UserDefinedDagsterK8sConfig.from_dict(
             user_defined_k8s_config_dict
         )
         check.opt_inst_param(
-            user_defined_k8s_config, "user_defined_k8s_config", UserDefinedDagsterK8sConfig,
+            user_defined_k8s_config,
+            "user_defined_k8s_config",
+            UserDefinedDagsterK8sConfig,
         )
         check.opt_str_param(kubeconfig_file, "kubeconfig_file")
 
@@ -283,12 +324,15 @@ def create_k8s_job_task(celery_app, **task_kwargs):
         else:
             kubernetes.config.load_kube_config(kubeconfig_file)
 
-        instance_ref = InstanceRef.from_dict(instance_ref_dict)
-        instance = DagsterInstance.from_ref(instance_ref)
-        pipeline_run = instance.get_run_by_id(run_id)
+        instance = DagsterInstance.from_ref(execute_step_args.instance_ref)
+        pipeline_run = instance.get_run_by_id(execute_step_args.pipeline_run_id)
 
-        check.invariant(pipeline_run, "Could not load run {}".format(run_id))
-        step_key = step_keys[0]
+        check.inst(
+            pipeline_run,
+            PipelineRun,
+            "Could not load run {}".format(execute_step_args.pipeline_run_id),
+        )
+        step_key = execute_step_args.step_keys_to_execute[0]
 
         celery_worker_name = self.request.hostname
         celery_pod_name = os.environ.get("HOSTNAME")
@@ -307,43 +351,36 @@ def create_k8s_job_task(celery_app, **task_kwargs):
 
         if pipeline_run.status != PipelineRunStatus.STARTED:
             instance.report_engine_event(
-                "Not scheduling step because pipeline run status is not STARTED",
+                "Not scheduling step because dagster run status is not STARTED",
                 pipeline_run,
-                EngineEventData([EventMetadataEntry.text(step_key, "Step keys"),]),
+                EngineEventData(
+                    [
+                        EventMetadataEntry.text(step_key, "Step key"),
+                    ]
+                ),
                 CeleryK8sJobExecutor,
                 step_key=step_key,
             )
-            return
+            return []
 
         # Ensure we stay below k8s name length limits
-        k8s_name_key = get_k8s_job_name(run_id, step_key)
+        k8s_name_key = get_k8s_job_name(execute_step_args.pipeline_run_id, step_key)
 
-        retries = Retries.from_config(retries_dict)
+        retry_state = execute_step_args.known_state.get_retry_state()
 
-        if retries.get_attempt_count(step_key):
-            attempt_number = retries.get_attempt_count(step_key)
+        if retry_state.get_attempt_count(step_key):
+            attempt_number = retry_state.get_attempt_count(step_key)
             job_name = "dagster-job-%s-%d" % (k8s_name_key, attempt_number)
             pod_name = "dagster-job-%s-%d" % (k8s_name_key, attempt_number)
         else:
             job_name = "dagster-job-%s" % (k8s_name_key)
             pod_name = "dagster-job-%s" % (k8s_name_key)
 
-        input_json = serialize_dagster_namedtuple(
-            ExecuteStepArgs(
-                pipeline_origin=pipeline_origin,
-                pipeline_run_id=run_id,
-                instance_ref=None,
-                mode=mode,
-                step_keys_to_execute=step_keys,
-                run_config=run_config,
-                retries_dict=retries_dict,
-            )
-        )
-        command = ["dagster"]
-        args = ["api", "execute_step_with_structured_logs", input_json]
+        input_json = serialize_dagster_namedtuple(execute_step_args)
+        args = ["dagster", "api", "execute_step", input_json]
 
         job = construct_dagster_k8s_job(
-            job_config, command, args, job_name, user_defined_k8s_config, pod_name
+            job_config, args, job_name, user_defined_k8s_config, pod_name
         )
 
         # Running list of events generated from this task execution
@@ -356,9 +393,8 @@ def create_k8s_job_task(celery_app, **task_kwargs):
             pipeline_run,
             EngineEventData(
                 [
-                    EventMetadataEntry.text(step_key, "Step keys"),
+                    EventMetadataEntry.text(step_key, "Step key"),
                     EventMetadataEntry.text(job_name, "Kubernetes Job name"),
-                    EventMetadataEntry.text(pod_name, "Kubernetes Pod name"),
                     EventMetadataEntry.text(job_config.job_image, "Job image"),
                     EventMetadataEntry.text(job_config.image_pull_policy, "Image pull policy"),
                     EventMetadataEntry.text(
@@ -376,21 +412,19 @@ def create_k8s_job_task(celery_app, **task_kwargs):
             step_key=step_key,
         )
         events.append(engine_event)
-
         try:
             kubernetes.client.BatchV1Api().create_namespaced_job(body=job, namespace=job_namespace)
         except kubernetes.client.rest.ApiException as e:
             if e.reason == "Conflict":
-                # There is an existing job with the same name so do not procede.
+                # There is an existing job with the same name so proceed and see if the existing job succeeded
                 instance.report_engine_event(
                     "Did not create Kubernetes job {} for step {} since job name already "
-                    "exists, exiting.".format(job_name, step_key),
+                    "exists, proceeding with existing job.".format(job_name, step_key),
                     pipeline_run,
                     EngineEventData(
                         [
-                            EventMetadataEntry.text(step_key, "Step keys"),
+                            EventMetadataEntry.text(step_key, "Step key"),
                             EventMetadataEntry.text(job_name, "Kubernetes Job name"),
-                            EventMetadataEntry.text(pod_name, "Kubernetes Pod name"),
                         ],
                         marker_end=DELEGATE_MARKER,
                     ),
@@ -404,26 +438,35 @@ def create_k8s_job_task(celery_app, **task_kwargs):
                     pipeline_run,
                     EngineEventData(
                         [
-                            EventMetadataEntry.text(step_key, "Step keys"),
-                            EventMetadataEntry.text(e, "Error"),
-                        ]
+                            EventMetadataEntry.text(step_key, "Step key"),
+                        ],
+                        error=serializable_error_info_from_exc_info(sys.exc_info()),
                     ),
                     CeleryK8sJobExecutor,
                     step_key=step_key,
                 )
-            return
+                return []
 
         try:
             wait_for_job_success(
-                job_name=job_name, namespace=job_namespace, instance=instance, run_id=run_id,
+                job_name=job_name,
+                namespace=job_namespace,
+                instance=instance,
+                run_id=execute_step_args.pipeline_run_id,
+                wait_timeout=job_wait_timeout,
             )
+        except (DagsterK8sError, DagsterK8sTimeoutError) as err:
+            step_failure_event = construct_step_failure_event_and_handle(
+                pipeline_run, step_key, err, instance=instance
+            )
+            events.append(step_failure_event)
         except DagsterK8sPipelineStatusException:
             instance.report_engine_event(
-                "Terminating Kubernetes Job because pipeline run status is not STARTED",
+                "Terminating Kubernetes Job because dagster run status is not STARTED",
                 pipeline_run,
                 EngineEventData(
                     [
-                        EventMetadataEntry.text(step_key, "Step keys"),
+                        EventMetadataEntry.text(step_key, "Step key"),
                         EventMetadataEntry.text(job_name, "Kubernetes Job name"),
                         EventMetadataEntry.text(job_namespace, "Kubernetes Job namespace"),
                     ]
@@ -432,9 +475,47 @@ def create_k8s_job_task(celery_app, **task_kwargs):
                 step_key=step_key,
             )
             delete_job(job_name=job_name, namespace=job_namespace)
-            return
+            return []
+        except (
+            DagsterK8sUnrecoverableAPIError,
+            DagsterK8sAPIRetryLimitExceeded,
+            # We shouldn't see unwrapped APIExceptions anymore, as they should all be wrapped in
+            # a retry boundary. We still catch it here just in case we missed one so that we can
+            # report it to the event log
+            kubernetes.client.rest.ApiException,
+        ) as err:
+            instance.report_engine_event(
+                "Encountered unexpected error while waiting on Kubernetes job {} for step {}, "
+                "exiting.".format(job_name, step_key),
+                pipeline_run,
+                EngineEventData(
+                    [
+                        EventMetadataEntry.text(step_key, "Step key"),
+                    ],
+                    error=serializable_error_info_from_exc_info(sys.exc_info()),
+                ),
+                CeleryK8sJobExecutor,
+                step_key=step_key,
+            )
+            return []
 
-        pod_names = get_pod_names_in_job(job_name, namespace=job_namespace)
+        try:
+            pod_names = get_pod_names_in_job(job_name, namespace=job_namespace)
+        except kubernetes.client.rest.ApiException as e:
+            instance.report_engine_event(
+                "Encountered unexpected error retreiving Pods for Kubernetes job {} for step {}, "
+                "exiting.".format(job_name, step_key),
+                pipeline_run,
+                EngineEventData(
+                    [
+                        EventMetadataEntry.text(step_key, "Step key"),
+                    ],
+                    error=serializable_error_info_from_exc_info(sys.exc_info()),
+                ),
+                CeleryK8sJobExecutor,
+                step_key=step_key,
+            )
+            return []
 
         # Post engine event for log retrieval
         engine_event = instance.report_engine_event(
@@ -448,8 +529,25 @@ def create_k8s_job_task(celery_app, **task_kwargs):
 
         logs = []
         for pod_name in pod_names:
-            raw_logs = retrieve_pod_logs(pod_name, namespace=job_namespace)
-            logs += raw_logs.split("\n")
+            try:
+                raw_logs = retrieve_pod_logs(pod_name, namespace=job_namespace)
+                logs += raw_logs.split("\n")
+            except kubernetes.client.rest.ApiException as e:
+                instance.report_engine_event(
+                    "Encountered unexpected error while fetching pod logs for Kubernetes job {}, "
+                    "Pod name {} for step {}. Will attempt to continue with other pods.".format(
+                        job_name, pod_name, step_key
+                    ),
+                    pipeline_run,
+                    EngineEventData(
+                        [
+                            EventMetadataEntry.text(step_key, "Step key"),
+                        ],
+                        error=serializable_error_info_from_exc_info(sys.exc_info()),
+                    ),
+                    CeleryK8sJobExecutor,
+                    step_key=step_key,
+                )
 
         events += filter_dagster_events_from_pod_logs(logs)
         serialized_events = [serialize_dagster_namedtuple(event) for event in events]
