@@ -1,6 +1,17 @@
-import weakref
+import sys
 
 import kubernetes
+from dagster import DagsterInvariantViolationError, EventMetadataEntry, Field, Noneable, check
+from dagster.config.field import resolve_to_config_type
+from dagster.config.validate import process_config
+from dagster.core.events import EngineEventData
+from dagster.core.execution.retries import RetryMode
+from dagster.core.launcher import LaunchRunContext, RunLauncher
+from dagster.core.storage.pipeline_run import PipelineRunStatus
+from dagster.core.storage.tags import DOCKER_IMAGE_TAG
+from dagster.serdes import ConfigurableClass, ConfigurableClassData, serialize_dagster_namedtuple
+from dagster.utils import frozentags, merge_dicts
+from dagster.utils.error import serializable_error_info_from_exc_info
 from dagster_k8s.job import (
     DagsterK8sJobConfig,
     construct_dagster_k8s_job,
@@ -9,31 +20,17 @@ from dagster_k8s.job import (
 )
 from dagster_k8s.utils import delete_job
 
-from dagster import DagsterInvariantViolationError, EventMetadataEntry, Field, Noneable, check
-from dagster.config.field import resolve_to_config_type
-from dagster.config.validate import process_config
-from dagster.core.events import EngineEventData
-from dagster.core.execution.retries import Retries
-from dagster.core.host_representation import ExternalPipeline
-from dagster.core.host_representation.handle import GrpcServerRepositoryLocationHandle
-from dagster.core.instance import DagsterInstance
-from dagster.core.launcher import RunLauncher
-from dagster.core.origin import PipelineGrpcServerOrigin, PipelinePythonOrigin
-from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus
-from dagster.serdes import ConfigurableClass, ConfigurableClassData, serialize_dagster_namedtuple
-from dagster.utils import frozentags, merge_dicts
-
 from .config import CELERY_K8S_CONFIG_KEY, celery_k8s_config
 
 
 class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
-    """In contrast to the :py:class:`K8sRunLauncher`, which launches pipeline runs as single K8s
+    """In contrast to the :py:class:`K8sRunLauncher`, which launches dagster runs as single K8s
     Jobs, this run launcher is intended for use in concert with
     :py:func:`dagster_celery_k8s.celery_k8s_job_executor`.
 
     With this run launcher, execution is delegated to:
 
-        1. A run coordinator Kubernetes Job, which traverses the pipeline run execution plan and
+        1. A run worker Kubernetes Job, which traverses the dagster run execution plan and
            submits steps to Celery queues for execution;
         2. The step executions which are submitted to Celery queues are picked up by Celery workers,
            and each step execution spawns a step execution Kubernetes Job. See the implementation
@@ -95,6 +92,7 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
         config_source=None,
         retries=None,
         inst_data=None,
+        k8s_client_batch_api=None,
     ):
         self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
 
@@ -108,6 +106,8 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
             check.opt_str_param(kubeconfig_file, "kubeconfig_file")
             kubernetes.config.load_kube_config(kubeconfig_file)
 
+        self._fixed_batch_api = k8s_client_batch_api
+
         self.instance_config_map = check.str_param(instance_config_map, "instance_config_map")
         self.dagster_home = check.str_param(dagster_home, "dagster_home")
         self.postgres_password_secret = check.str_param(
@@ -119,8 +119,12 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
         self.config_source = check.opt_dict_param(config_source, "config_source")
 
         retries = check.opt_dict_param(retries, "retries") or {"enabled": {}}
-        self.retries = Retries.from_config(retries)
-        self._instance_ref = None
+        self.retries = RetryMode.from_config(retries)
+        super().__init__()
+
+    @property
+    def _batch_api(self):
+        return self._fixed_batch_api if self._fixed_batch_api else kubernetes.client.BatchV1Api()
 
     @classmethod
     def config_type(cls):
@@ -147,64 +151,44 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
     def inst_data(self):
         return self._inst_data
 
-    @property
-    def _instance(self):
-        return self._instance_ref() if self._instance_ref else None
-
-    def initialize(self, instance):
-        check.inst_param(instance, "instance", DagsterInstance)
-        # Store a weakref to avoid a circular reference / enable GC
-        self._instance_ref = weakref.ref(instance)
-
-    def launch_run(self, instance, run, external_pipeline):
-        check.inst_param(instance, "instance", DagsterInstance)
-        check.inst_param(run, "run", PipelineRun)
-        check.inst_param(external_pipeline, "external_pipeline", ExternalPipeline)
+    def launch_run(self, context: LaunchRunContext) -> None:
+        run = context.pipeline_run
 
         job_name = get_job_name_from_run_id(run.run_id)
         pod_name = job_name
         exc_config = _get_validated_celery_k8s_executor_config(run.run_config)
-
-        job_image = None
-        pipeline_origin = None
         env_vars = None
-        if isinstance(external_pipeline.get_origin(), PipelineGrpcServerOrigin):
-            if exc_config.get("job_image"):
-                raise DagsterInvariantViolationError(
-                    "Cannot specify job_image in executor config when loading pipeline "
-                    "from GRPC server."
+
+        job_image_from_executor_config = exc_config.get("job_image")
+
+        pipeline_origin = context.pipeline_code_origin
+        repository_origin = pipeline_origin.repository_origin
+
+        job_image = repository_origin.container_image
+
+        if job_image:
+            if job_image_from_executor_config:
+                job_image = job_image_from_executor_config
+                self._instance.report_engine_event(
+                    f"You have specified a job_image {job_image_from_executor_config} in your executor configuration, "
+                    f"but also {job_image} in your user-code deployment. Using the job image {job_image_from_executor_config} "
+                    f"from executor configuration as it takes precedence.",
+                    run,
+                    cls=self.__class__,
                 )
-
-            repository_location_handle = (
-                external_pipeline.repository_handle.repository_location_handle
-            )
-
-            if not isinstance(repository_location_handle, GrpcServerRepositoryLocationHandle):
-                raise DagsterInvariantViolationError(
-                    "Expected RepositoryLocationHandle to be of type "
-                    "GrpcServerRepositoryLocationHandle but found type {}".format(
-                        type(repository_location_handle)
-                    )
-                )
-
-            job_image = repository_location_handle.get_current_image()
-            env_vars = {"DAGSTER_CURRENT_IMAGE": job_image}
-
-            repository_name = external_pipeline.repository_handle.repository_name
-            pipeline_origin = PipelinePythonOrigin(
-                pipeline_name=external_pipeline.name,
-                repository_origin=repository_location_handle.get_repository_python_origin(
-                    repository_name
-                ),
-            )
-
         else:
-            job_image = exc_config.get("job_image")
-            if not job_image:
+            if not job_image_from_executor_config:
                 raise DagsterInvariantViolationError(
-                    "Cannot find job_image in celery-k8s executor config."
+                    "You have not specified a job_image in your executor configuration. "
+                    "To resolve this error, specify the job_image configuration in the executor "
+                    "config section in your run config. \n"
+                    "Note: You may also be seeing this error because you are using the configured API. "
+                    "Using configured with the celery-k8s executor is not supported at this time, "
+                    "and the job_image must be configured at the top-level executor config without "
+                    "using configured."
                 )
-            pipeline_origin = external_pipeline.get_origin()
+
+            job_image = job_image_from_executor_config
 
         job_config = DagsterK8sJobConfig(
             dagster_home=self.dagster_home,
@@ -216,9 +200,16 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
             service_account_name=exc_config.get("service_account_name"),
             env_config_maps=exc_config.get("env_config_maps"),
             env_secrets=exc_config.get("env_secrets"),
+            volume_mounts=exc_config.get("volume_mounts"),
+            volumes=exc_config.get("volumes"),
         )
 
-        user_defined_k8s_config = get_user_defined_k8s_config(frozentags(external_pipeline.tags))
+        self._instance.add_run_tags(
+            run.run_id,
+            {DOCKER_IMAGE_TAG: job_config.job_image},
+        )
+
+        user_defined_k8s_config = get_user_defined_k8s_config(frozentags(run.tags))
 
         from dagster.cli.api import ExecuteRunArgs
 
@@ -226,14 +217,15 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
             # depends on DagsterInstance.get() returning the same instance
             # https://github.com/dagster-io/dagster/issues/2757
             ExecuteRunArgs(
-                pipeline_origin=pipeline_origin, pipeline_run_id=run.run_id, instance_ref=None,
+                pipeline_origin=pipeline_origin,
+                pipeline_run_id=run.run_id,
+                instance_ref=None,
             )
         )
 
         job = construct_dagster_k8s_job(
             job_config,
-            command=["dagster"],
-            args=["api", "execute_run_with_structured_logs", input_json],
+            args=["dagster", "api", "execute_run", input_json],
             job_name=job_name,
             pod_name=pod_name,
             component="run_coordinator",
@@ -243,23 +235,32 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
 
         job_namespace = exc_config.get("job_namespace")
 
-        api = kubernetes.client.BatchV1Api()
-        api.create_namespaced_job(body=job, namespace=job_namespace)
-
         self._instance.report_engine_event(
-            "Kubernetes run_coordinator job launched",
+            "Creating Kubernetes run worker job",
             run,
             EngineEventData(
                 [
                     EventMetadataEntry.text(job_name, "Kubernetes Job name"),
-                    EventMetadataEntry.text(pod_name, "Kubernetes Pod name"),
                     EventMetadataEntry.text(job_namespace, "Kubernetes Namespace"),
                     EventMetadataEntry.text(run.run_id, "Run ID"),
                 ]
             ),
-            cls=CeleryK8sRunLauncher,
+            cls=self.__class__,
+        )
+
+        self._batch_api.create_namespaced_job(body=job, namespace=job_namespace)
+        self._instance.report_engine_event(
+            "Kubernetes run worker job created",
+            run,
+            EngineEventData(
+                [
+                    EventMetadataEntry.text(job_name, "Kubernetes Job name"),
+                    EventMetadataEntry.text(job_namespace, "Kubernetes Namespace"),
+                    EventMetadataEntry.text(run.run_id, "Run ID"),
+                ]
+            ),
+            cls=self.__class__,
         )
-        return run
 
     # https://github.com/dagster-io/dagster/issues/2741
     def can_terminate(self, run_id):
@@ -277,14 +278,53 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
     def terminate(self, run_id):
         check.str_param(run_id, "run_id")
 
-        if not self.can_terminate(run_id):
+        run = self._instance.get_run_by_id(run_id)
+        if not run:
+            return False
+
+        can_terminate = self.can_terminate(run_id)
+        if not can_terminate:
+            self._instance.report_engine_event(
+                message="Unable to terminate dagster job: can_terminate returned {}.".format(
+                    can_terminate
+                ),
+                pipeline_run=run,
+                cls=self.__class__,
+            )
             return False
 
         job_name = get_job_name_from_run_id(run_id)
 
         job_namespace = self.get_namespace_from_run_config(run_id)
 
-        return delete_job(job_name=job_name, namespace=job_namespace)
+        self._instance.report_run_canceling(run)
+
+        try:
+            termination_result = delete_job(job_name=job_name, namespace=job_namespace)
+            if termination_result:
+                self._instance.report_engine_event(
+                    message="Dagster Job was terminated successfully.",
+                    pipeline_run=run,
+                    cls=self.__class__,
+                )
+            else:
+                self._instance.report_engine_event(
+                    message="Dagster Job was not terminated successfully; delete_job returned {}".format(
+                        termination_result
+                    ),
+                    pipeline_run=run,
+                    cls=self.__class__,
+                )
+            return termination_result
+        except Exception:  # pylint: disable=broad-except
+            self._instance.report_engine_event(
+                message="Dagster Job was not terminated successfully; encountered error in delete_job",
+                pipeline_run=run,
+                engine_event_data=EngineEventData.engine_error(
+                    serializable_error_info_from_exc_info(sys.exc_info())
+                ),
+                cls=self.__class__,
+            )
 
     def get_namespace_from_run_config(self, run_id):
         check.str_param(run_id, "run_id")
@@ -298,14 +338,19 @@ class CeleryK8sRunLauncher(RunLauncher, ConfigurableClass):
 def _get_validated_celery_k8s_executor_config(run_config):
     check.dict_param(run_config, "run_config")
 
-    check.invariant(
-        CELERY_K8S_CONFIG_KEY in run_config.get("execution", {}),
-        "{} execution must be configured in pipeline execution config to launch runs with "
-        "CeleryK8sRunLauncher".format(CELERY_K8S_CONFIG_KEY),
-    )
+    executor_config = run_config.get("execution", {})
+    if not CELERY_K8S_CONFIG_KEY in executor_config:
+        raise DagsterInvariantViolationError(
+            "{config_key} execution configuration must be present in the run config to use the CeleryK8sRunLauncher. "
+            "Note: You may also be seeing this error because you are using the configured API. "
+            "Using configured with the {config_key} executor is not supported at this time, "
+            "and all executor config must be directly in the run config without using configured.".format(
+                config_key=CELERY_K8S_CONFIG_KEY,
+            ),
+        )
 
     execution_config_schema = resolve_to_config_type(celery_k8s_config())
-    execution_run_config = run_config["execution"][CELERY_K8S_CONFIG_KEY].get("config", {})
+    execution_run_config = (run_config["execution"][CELERY_K8S_CONFIG_KEY] or {}).get("config", {})
     res = process_config(execution_config_schema, execution_run_config)
 
     check.invariant(
